""" Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung) """
import uuid
from unittest.mock import AsyncMock

import inject
import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.evaluator_factory import create_and_inject_hardcoded_evaluators
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.evaluation_results import FormatConstraintEvaluationResult, RequirementConstraintEvaluationResult
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.expressions.enums import ModalMark, PrefixOperator, RequirementIndicator
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.expressions.hints_provider import HintsProvider

pytestmark = pytest.mark.asyncio


class TestAHBExpressionEvaluation:
    """Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung)"""

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(HintsProvider, AsyncMock(wraps=HintsProvider)).bind(
                RcEvaluator, AsyncMock(wraps=RcEvaluator)
            )
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        """ahb_expression, expected_requirement_indicator, expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional, expected_format_constraints_expression, expected_hints""",
        [
            pytest.param("Muss", ModalMark.MUSS, True, False, None, None),
            pytest.param("X", PrefixOperator.X, True, False, None, None),
            pytest.param("Muss[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("muss[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("M[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("m[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Soll[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("soll[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("S[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("s[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("Kann[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("kann[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("K[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("k[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("Muss[2]", ModalMark.MUSS, False, True, None, None),
            pytest.param("Muss[1]Soll[2]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Soll[1]Kann[4]", ModalMark.SOLL, True, True, None, None),
            pytest.param("Muss[2]\tSoll [ 1]  Kann[4\t]", ModalMark.SOLL, True, True, None, None),
            pytest.param("X[1]", PrefixOperator.X, True, True, None, None),
            pytest.param("O[2]", PrefixOperator.O, False, True, None, None),
            pytest.param("U[1]", PrefixOperator.U, True, True, None, None),
            pytest.param("U[1]U[2]", PrefixOperator.U, False, True, None, None),
            pytest.param("X[1]U[2]U[3]O[4]", PrefixOperator.X, False, True, None, None),
            pytest.param("Muss([1]O[2])U[3]Soll[2]Kann[2]O[4]", ModalMark.MUSS, True, True, None, None),
            pytest.param("muss([1]o[2])u[3]soll[2]kann[2]O[4]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Soll[2]", ModalMark.SOLL, False, True, None, None),
            # Neutral value
            pytest.param("Muss[503]", ModalMark.MUSS, True, False, None, "[503]"),
            pytest.param("Soll[902]", ModalMark.SOLL, True, False, "[902]", None),
            pytest.param("Kann[503]U[504]", ModalMark.KANN, True, False, None, "[503]U[504]"),
            # Last modal mark without conditions
            pytest.param("Muss[1]Kann", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Kann", ModalMark.KANN, True, True, None, None),
        ],
    )
    async def test_evaluate_valid_ahb_expression(
        self,
        mocker,
        ahb_expression,
        expected_requirement_indicator: RequirementIndicator,
        expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional,
        expected_format_constraints_expression,
        expected_hints,
    ):
        """
        Tests that valid ahb expressions are evaluated as expected.
        Odd condition_keys are True, even condition_keys are False
        """

        def side_effect_rc_evaluation(condition_expression):
            if condition_expression.lower() in ["[1]", " [ 1]  ", "[3]", "([1]o[2])u[3]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression.lower() in [
                "[2]",
                "[2]\t",
                "[4]",
                "[4\t]",
                "[1]u[2]",
                "[1]u[2]u[3]o[4]",
                "[2]o[4]",
            ]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=False,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression.lower() in ["[503]", "[503]u[504]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=None,
                    hints=condition_expression,
                )
            if condition_expression.lower() in ["[902]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=condition_expression,
                    hints=None,
                )

        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.requirement_constraint_evaluation",
            side_effect=AsyncMock(side_effect=side_effect_rc_evaluation),
        )
        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.format_constraint_evaluation",
            return_value=AsyncMock(
                side_effect=FormatConstraintEvaluationResult(format_constraints_fulfilled=True, error_message=None)
            ),
        )

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        result = await evaluate_ahb_expression_tree(parsed_tree, entered_input=None)  # type:ignore[arg-type]

        assert result.requirement_indicator == expected_requirement_indicator
        assert (
            result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            == expected_requirement_constraints_fulfilled
        )
        assert (
            result.requirement_constraint_evaluation_result.requirement_is_conditional
            == expected_requirement_is_conditional
        )
        assert (
            result.requirement_constraint_evaluation_result.format_constraints_expression
            == expected_format_constraints_expression
        )
        assert result.requirement_constraint_evaluation_result.hints == expected_hints

    @pytest.mark.parametrize(
        "expression, expected_error, expected_error_message",
        [
            pytest.param(
                "Muss[]",
                SyntaxError,
                """Please make sure that:
             * all conditions have the form [INT]
             * all packages have the form [INTPn..m]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """,
            ),
            pytest.param(
                "Soll[1001]",
                ValueError,
                """Condition key is not in valid number range.""",
            ),
        ],
    )
    async def test_invalid_ahb_expressions(
        self, expression: str, expected_error: type, expected_error_message: str, setup_and_teardown_injector
    ):
        """Tests that an error is raised when trying to pass invalid values."""

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)

        with pytest.raises(expected_error) as excinfo:  # type: ignore[var-annotated]
            await evaluate_ahb_expression_tree(
                parsed_tree, entered_input=None  # type:ignore[arg-type] # ok because error test
            )

        assert expected_error_message in str(excinfo.value)

    @pytest.mark.parametrize(
        "ahb_expression, content_evaluation_result",
        [
            pytest.param(
                "Muss ([2]O[3])[902][501]",
                ContentEvaluationResult(
                    hints={"501": "foo"},
                    format_constraints={
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "2": ConditionFulfilledValue.FULFILLED,
                        "3": ConditionFulfilledValue.UNFULFILLED,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                    packages={},
                ),
            )
        ],
    )
    async def test_all_serializations_work_similar(
        self, ahb_expression: str, content_evaluation_result: ContentEvaluationResult
    ):
        tree_a = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        tree_b = await parse_expression_including_unresolved_subexpressions(ahb_expression)
        # it's OK/expected that the trees look different depending on whether sub expressions are resolved or not
        # but in any case the evaluation result should look the same
        create_and_inject_hardcoded_evaluators(content_evaluation_result)
        evaluation_input = "something has to be here but it's not important what"
        evaluation_result_a = await evaluate_ahb_expression_tree(tree_a, entered_input=evaluation_input)
        evaluation_result_b = await evaluate_ahb_expression_tree(tree_b, entered_input=evaluation_input)
        assert evaluation_result_a == evaluation_result_b



"""
Tests the utility functions.
"""
from typing import Awaitable, List, TypeVar, Union

import pytest  # type:ignore[import]

from ahbicht.mapping_results import Repeatability, parse_repeatability
from ahbicht.utility_functions import gather_if_necessary

pytestmark = pytest.mark.asyncio

T = TypeVar("T")


async def _return_awaitable(arg: T) -> T:
    """returns the argument but is awaitable"""
    return arg


class TestUtilityFunctions:
    @pytest.mark.parametrize(
        "mixed_input, expected_result",
        [
            # only non-awaitables
            pytest.param([], []),
            pytest.param([3, 2, 1], [3, 2, 1]),
            pytest.param(["a", "b", "c"], ["a", "b", "c"]),
            # both awaitables and non-awaitables
            pytest.param(
                ["2", _return_awaitable("3"), "4", "foo", _return_awaitable("bar")], ["2", "3", "4", "foo", "bar"]
            ),
            # only awaitables
            pytest.param([_return_awaitable("a"), _return_awaitable("b")], ["a", "b"]),
        ],
    )
    async def test_gather_if_necessary(self, mixed_input: List[Union[T, Awaitable[T]]], expected_result: List[T]):
        actual = await gather_if_necessary(mixed_input)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "candidate, expected_result",
        [
            pytest.param("0..1", Repeatability(min_occurrences=0, max_occurrences=1)),
            pytest.param("1..1", Repeatability(min_occurrences=1, max_occurrences=1)),
            pytest.param("1..2", Repeatability(min_occurrences=1, max_occurrences=2)),
            pytest.param("71..89", Repeatability(min_occurrences=71, max_occurrences=89)),
        ],
    )
    def test_parse_repeatability(self, candidate: str, expected_result: Repeatability):
        actual = parse_repeatability(candidate)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "invalid_candidate",
        [
            pytest.param("1..0"),
            pytest.param("77..76"),
            pytest.param("0..0"),
            pytest.param("-1..1"),
            pytest.param("bullshit"),
            pytest.param(""),
        ],
    )
    def test_parse_repeatability_failures(self, invalid_candidate):
        with pytest.raises(ValueError):
            _ = parse_repeatability(invalid_candidate)



"""
Module to create expressions from scratch.
"""
import re
from abc import ABC, abstractmethod
from typing import Generic, Optional, Protocol, Type, TypeVar, Union

from ahbicht.expressions.condition_nodes import (
    ConditionNode,
    EvaluatedComposition,
    EvaluatedFormatConstraint,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.enums import LogicalOperator

TSupportedNodes = TypeVar("TSupportedNodes")


class ExpressionBuilder(Generic[TSupportedNodes], ABC):
    """
    Class that helps to create expression strings. It separates the logical operation (connect two conditions with a
    logical operator) from the implementation which might differ depending on the condition type and other
    circumstances.
    """

    @abstractmethod
    def get_expression(self) -> Optional[str]:
        """
        Returns the expression string or none if there is no expression.
        :return:
        """
        raise NotImplementedError("Has to be implemented by inheriting class.")

    @abstractmethod
    def land(self, other: TSupportedNodes):
        """
        connects the expression with a logical and (LAND)
        :param other: condition or expression to be connected to the expression
        :return:
        """
        raise NotImplementedError("Has to be implemented by inheriting class.")

    @abstractmethod
    def lor(self, other: TSupportedNodes):
        """
        connects the expression with a logical or (LOR)
        :param other: condition or expression to be connected to the expression
        :return:
        """
        raise NotImplementedError("Has to be implemented by inheriting class.")

    @abstractmethod
    def xor(self, other: TSupportedNodes):
        """
        connects the expression with an exclusive or (XOR)
        :param other: condition or expression to be connected to the expression
        :return:
        """
        raise NotImplementedError("Has to be implemented by inheriting class.")


TEffectiveFCExpressionBuilderArguments = Union[
    EvaluatedComposition, UnevaluatedFormatConstraint, Optional[str]
]  # node types that have an effect on the built format constraint expression

TUneffectiveFCExpressionBuilderArguments = Union[
    RequirementConstraint, EvaluatedComposition, Hint, Type[ConditionNode]
]  # node types that are formally accepted as argument but don't
# have any effect. Instead of checking which nodes contain format constraints all are put into the
# FormatConstraintExpressionBuilder, but it only has an effect on those with format constraints.
# Note that EvaluatedComposition is in both classes since they can have format constraints but don't have to.

TSupportedFCExpressionBuilderArguments = Union[
    TEffectiveFCExpressionBuilderArguments, TUneffectiveFCExpressionBuilderArguments
]


class FormatConstraintExpressionBuilder(ExpressionBuilder[TSupportedFCExpressionBuilderArguments]):
    """
    Class to create expressions that consists of FormatConstraints
    """

    _one_key_surrounded_by_brackets_pattern = re.compile(r"\((?P<body>\[\d+\])\)")  # https://regex101.com/r/IauOei/1

    # (?P<group_name>...) is a named group: https://docs.python.org/3/howto/regex.html#non-capturing-and-named-groups

    def __init__(self, init_condition_or_expression: TSupportedFCExpressionBuilderArguments):
        """
        Start with a plain expression
        :param init_condition_or_expression: initial format constraint or existing expression
        """
        self._expression: Optional[str]
        if isinstance(init_condition_or_expression, UnevaluatedFormatConstraint):
            # the condition key of the token in expression '[42]' is only '42'
            # so the get a valid expression, we add the square brackets
            self._expression = f"[{init_condition_or_expression.condition_key}]"
        elif (
            isinstance(init_condition_or_expression, EvaluatedComposition)
            and init_condition_or_expression.format_constraints_expression
        ):
            self._expression = init_condition_or_expression.format_constraints_expression
        elif isinstance(init_condition_or_expression, str):
            self._expression = f"{init_condition_or_expression}"
        elif isinstance(init_condition_or_expression, (RequirementConstraint, EvaluatedComposition, Hint)):
            # requirement constraints and hints don't have any effect on the a newly built
            # format constraint expression
            # also evaluated compositions that don't have a format constraint expression
            self._expression = None
        else:
            # we should never come here
            self._expression = None

    def get_expression(self) -> Optional[str]:
        # could add simplifications here
        return self._expression

    def land(self, other: TSupportedFCExpressionBuilderArguments) -> ExpressionBuilder:
        return self._connect(LogicalOperator.LAND, other)

    def lor(self, other: TSupportedFCExpressionBuilderArguments) -> ExpressionBuilder:
        return self._connect(LogicalOperator.LOR, other)

    def xor(self, other: TSupportedFCExpressionBuilderArguments) -> ExpressionBuilder:
        return self._connect(LogicalOperator.XOR, other)

    def _connect(self, operator_character: LogicalOperator, other: TSupportedFCExpressionBuilderArguments):
        """
        Connect the existing expression and the other part.

        :param operator_character: "X", "U" or "O"
        """
        if self._expression:
            prefix = f"({self._expression}) {operator_character}"
        else:
            prefix = ""
        if isinstance(other, UnevaluatedFormatConstraint):
            self._expression = f"{prefix} [{other.condition_key}]"
        elif isinstance(other, EvaluatedComposition) and other.format_constraints_expression:
            self._expression = f"{prefix} ({other.format_constraints_expression})"
        elif isinstance(other, str):
            self._expression = f"{prefix} ({other})"
        elif isinstance(other, (RequirementConstraint, EvaluatedComposition, Hint)):
            # other types than the above don't affect the newly built format constraint expression (no effect, explicit)
            pass
        else:
            # all other types also have no effect (no effect, implicit)
            pass  # we should never come here
        if self._expression:
            self._expression = self._expression.strip()
            self._expression = self._one_key_surrounded_by_brackets_pattern.sub(r"\g<body>", self._expression)
        return self


# pylint:disable=too-few-public-methods
class _ClassesWithHintAttribute(Protocol):
    """
    A class to be used in type hints. describes all classes that have a "hint" attribute
    """

    hint: str


TClassesWithHintAttribute = TypeVar("TClassesWithHintAttribute", bound=_ClassesWithHintAttribute)


class HintExpressionBuilder(ExpressionBuilder[TClassesWithHintAttribute]):
    """
    Allows to connect hints with logical operations.
    """

    @staticmethod
    def get_hint_text(hinty_object: Optional[_ClassesWithHintAttribute]) -> Optional[str]:
        """
        get the hint from a Hint instance or plain string
        :param hinty_object:
        :return: hint if there is any, None otherwise
        """
        if hinty_object is None:
            return None
        if isinstance(hinty_object, str):
            return hinty_object
        return getattr(hinty_object, "hint", None)

    def __init__(self, init_condition: Optional[_ClassesWithHintAttribute]):
        """
        Initialize by providing either a Hint Node or a hint string
        """
        self._expression = HintExpressionBuilder.get_hint_text(init_condition)

    def get_expression(self) -> Optional[str]:
        return self._expression

    def land(self, other: Optional[_ClassesWithHintAttribute]) -> ExpressionBuilder:
        if other is not None:
            if self._expression:
                self._expression += f" und {HintExpressionBuilder.get_hint_text(other)}"
            else:
                self._expression = HintExpressionBuilder.get_hint_text(other)
        return self

    def lor(self, other: Optional[_ClassesWithHintAttribute]) -> ExpressionBuilder:
        if other is not None:
            if self._expression:
                self._expression += f" oder {HintExpressionBuilder.get_hint_text(other)}"
            else:
                self._expression = HintExpressionBuilder.get_hint_text(other)
        return self

    def xor(self, other: Optional[_ClassesWithHintAttribute]) -> ExpressionBuilder:
        if other is not None:
            if self._expression:
                self._expression = f"Entweder ({self._expression}) oder ({HintExpressionBuilder.get_hint_text(other)})"
            else:
                self._expression = HintExpressionBuilder.get_hint_text(other)
        return self


# This class is only used by the FormatConstraintTransformer.
# That's why it only accepts EvaluatedFormatConstraints as input.
class FormatErrorMessageExpressionBuilder(ExpressionBuilder[EvaluatedFormatConstraint]):
    """
    Class to build the error messages for the format constraint evaluation.
    """

    def __init__(self, init_condition: EvaluatedFormatConstraint):
        self._expression = init_condition.error_message
        self.format_constraint_fulfilled = init_condition.format_constraint_fulfilled

    def get_expression(self) -> Optional[str]:
        return self._expression

    def land(self, other: EvaluatedFormatConstraint) -> ExpressionBuilder:
        if other.format_constraint_fulfilled is True:
            self._expression = self._expression
        else:
            if self._expression is None:
                self._expression = other.error_message
            else:
                self._expression = f"'{self._expression}' und '{other.error_message}'"
        return self

    def lor(self, other: EvaluatedFormatConstraint) -> ExpressionBuilder:
        if self.format_constraint_fulfilled is False and other.format_constraint_fulfilled is False:
            self._expression = f"'{self._expression}' oder '{other.error_message}'"
        else:
            self._expression = None
        return self

    def xor(self, other: EvaluatedFormatConstraint) -> ExpressionBuilder:
        if self.format_constraint_fulfilled is False and other.format_constraint_fulfilled is False:
            self._expression = f"Entweder '{self._expression}' oder '{other.error_message}'"
        elif self.format_constraint_fulfilled is True and other.format_constraint_fulfilled is True:
            self._expression = "Zwei exklusive Formatdefinitionen dürfen nicht gleichzeitig erfüllt sein"
            # TODO: Do we need to know which one? It's probably more work than benefit.
        else:
            self._expression = None
        return self



"""
This module makes it possible to parse expressions including all their subexpressions, if present.
for example ahb_expressions which contain condition_expressions or condition_expressions which contain packages.
Parsing expressions that are nested into other expressions is referred to as "resolving".
"""
import asyncio
import inspect
from typing import Awaitable, List, Optional, Union

import inject
from lark import Token, Transformer, Tree
from lark.exceptions import VisitError

from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.package_expansion import PackageResolver
from ahbicht.mapping_results import Repeatability, parse_repeatability


async def parse_expression_including_unresolved_subexpressions(
    expression: str, resolve_packages: bool = False
) -> Tree[Token]:
    """
    Parses expressions and resolves its subexpressions,
    for example condition_expressions in ahb_expressions or packages in condition_expressions.
    :param expression: a syntactically valid ahb_expression or condition_expression
    :param resolve_packages: if true resolves also the packages in the condition_expressions
    """
    try:
        expression_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)
        resolved_expression_tree = AhbExpressionResolverTransformer().transform(expression_tree)
        if resolve_packages:
            # the condition expression inside the ahb expression has to be resolved before trying to resolve packages
            resolved_expression_tree = await expand_packages(resolved_expression_tree)
    except SyntaxError as ahb_syntax_error:
        try:
            resolved_expression_tree = parse_condition_expression_to_tree(expression)
        except SyntaxError as condition_syntax_error:
            # pylint: disable=raise-missing-from
            raise SyntaxError(f"{ahb_syntax_error.msg} {condition_syntax_error.msg}")
    return resolved_expression_tree


async def expand_packages(parsed_tree: Tree) -> Tree[Token]:
    """
    Replaces all the "short" packages in parser_tree with the respective "long" condition expressions
    """
    try:
        result = PackageExpansionTransformer().transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc
    result = await _replace_sub_coroutines_with_awaited_results(result)
    return result


async def _replace_sub_coroutines_with_awaited_results(tree: Union[Tree, Awaitable[Tree]]) -> Tree[Token]:
    """
    awaits all coroutines inside the tree and replaces the coroutines with their respective awaited result.
    returns an updated tree
    """
    result: Tree
    if inspect.isawaitable(tree):
        result = await tree
    else:
        # if the tree type hint is correct this is always a tree if it's not awaitable
        result = tree  # type:ignore[assignment]
    # todo: check why lark type hints state the return value of scan_values is always Iterator[str]
    sub_results = await asyncio.gather(*result.scan_values(asyncio.iscoroutine))  # type:ignore[call-overload]
    for coro, sub_result in zip(result.scan_values(asyncio.iscoroutine), sub_results):
        for sub_tree in result.iter_subtrees():
            for child_index, child in enumerate(sub_tree.children):
                if child == coro:
                    sub_tree.children[child_index] = sub_result
    return result


# pylint: disable=no-self-use, invalid-name
class AhbExpressionResolverTransformer(Transformer):
    """
    Resolves the condition_expressions inside an ahb_expression.
    """

    def CONDITION_EXPRESSION(self, expression):
        """
        Replacing the expression_condition with its parsed tree.
        """
        condition_tree = parse_condition_expression_to_tree(expression.value)
        return condition_tree


# pylint: disable=no-self-use, invalid-name
class PackageExpansionTransformer(Transformer):
    """
    The PackageExpansionTransformer expands packages inside a tree to condition expressions by using a PackageResolver.
    """

    def __init__(self):
        super().__init__()
        self._resolver: PackageResolver = inject.instance(PackageResolver)

    def package(self, tokens: List[Token]) -> Awaitable[Tree]:
        """
        try to resolve the package using the injected PackageResolver
        """
        # The grammar guarantees that there is always exactly 1 package_key token/terminal.
        # But the repeatability token is optional, so the list repeatability_tokens might contain 0 or 1 entries
        # They all come in the same `tokens` list which we split in the following two lines.
        package_key_token = [t for t in tokens if t.type == "PACKAGE_KEY"][0]
        repeatability_tokens = [t for t in tokens if t.type == "REPEATABILITY"]
        # pylint: disable=unused-variable
        # we parse the repeatability, but we don't to anything with it, yet.
        repeatability: Optional[Repeatability]
        if len(repeatability_tokens) == 1:
            repeatability = parse_repeatability(repeatability_tokens[0].value)
        else:
            repeatability = None
        return self._package_async(package_key_token)

    async def _package_async(self, package_key_token: Token) -> Tree[Token]:
        resolved_package = await self._resolver.get_condition_expression(package_key_token.value)
        if not resolved_package.has_been_resolved_successfully():
            raise NotImplementedError(
                f"The package '{package_key_token.value}' could not be resolved by {self._resolver}"
            )
        # the package_expression is not None because that's the definition of "has been resolved successfully"
        tree_result = parse_condition_expression_to_tree(resolved_package.package_expression)  # type:ignore[arg-type]
        return tree_result



"""
This script is run in the tox 'json_schemas' environment.
It creates json schema files as described in the README.md in the same directory.
"""

import json
import pathlib
from typing import List, Type

from marshmallow import Schema, fields
from marshmallow_jsonschema import JSONSchema  # type:ignore[import]

from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtractSchema
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResultSchema
from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResultSchema,
    FormatConstraintEvaluationResultSchema,
    RequirementConstraintEvaluationResultSchema,
)
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraintSchema
from ahbicht.json_serialization.tree_schema import TokenSchema  # , TreeSchema
from ahbicht.mapping_results import ConditionKeyConditionTextMappingSchema, PackageKeyConditionExpressionMappingSchema

schema_types: List[Type[Schema]] = [
    RequirementConstraintEvaluationResultSchema,
    FormatConstraintEvaluationResultSchema,
    EvaluatedFormatConstraintSchema,
    AhbExpressionEvaluationResultSchema,
    ConditionKeyConditionTextMappingSchema,
    PackageKeyConditionExpressionMappingSchema,
    ContentEvaluationResultSchema,
    TokenSchema,
    CategorizedKeyExtractSchema,
    # TreeSchema
    # As of 2021-11 the TreeSchema fails, probably because of recursion or the lambda:
    # (<class 'AttributeError'>, AttributeError("'function' object has no attribute 'fields'"), ....)
]
json_schema = JSONSchema()
for schema_type in schema_types:
    this_directory = pathlib.Path(__file__).parent.absolute()
    file_name = schema_type.__name__ + ".json"  # pylint:disable=invalid-name
    file_path = this_directory / file_name
    schema_instance = schema_type()
    if "requirement_indicator" in schema_instance.fields:
        # workaround: in the schemas we want the requirement indicator to appear as simple string
        # the schema used internally is just a workaround
        for field_dict in [
            schema_instance.fields,
            schema_instance.load_fields,
            schema_instance.dump_fields,
            schema_instance.declared_fields,
        ]:
            field_dict["requirement_indicator"] = fields.String(name="requirement_indicator")

    json_schema_dict = json_schema.dump(schema_instance)

    with open(file_path, "w", encoding="utf-8") as json_schema_file:
        json.dump(json_schema_dict, json_schema_file, ensure_ascii=False, sort_keys=True, indent=4)



"""
This module evaluates the format constraint expression tree using EvaluatedFormatConstraints.
The FormatConstraintTransformer defines the rules how the different parts and nodes
of the format constraint expression tree are handled.

The used terms are defined in the README_conditions.md.
"""

from typing import Dict, List, Mapping, Optional

import inject
from lark import Token, Tree, v_args
from lark.exceptions import VisitError

from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.evaluation_results import FormatConstraintEvaluationResult
from ahbicht.expressions.base_transformer import BaseTransformer
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint
from ahbicht.expressions.expression_builder import FormatErrorMessageExpressionBuilder


# pylint: disable=no-self-use
@v_args(inline=True)  # Children are provided as *args instead of a list argument
# pylint:disable=inherit-non-class
class FormatConstraintTransformer(BaseTransformer[EvaluatedFormatConstraint, EvaluatedFormatConstraint]):
    """
    Transformer that evaluates the trees built from the format constraint expressions.
    The input are the evaluated format constraint conditions in the form of EvaluatedFormatConstraints.
    The return value is an EvaluatedFormatConstraint whose attribute `format_constraint_fulfilled` describes whether
    the format constraint expression is fulfilled or not.
    """

    def and_composition(
        self, left: EvaluatedFormatConstraint, right: EvaluatedFormatConstraint
    ) -> EvaluatedFormatConstraint:
        """Evaluates logical and_composition"""

        resulting_format_constraint_fulfilled = left.format_constraint_fulfilled and right.format_constraint_fulfilled
        error_message = FormatErrorMessageExpressionBuilder(left).land(right).get_expression()

        return EvaluatedFormatConstraint(
            format_constraint_fulfilled=resulting_format_constraint_fulfilled, error_message=error_message
        )

    def or_composition(
        self, left: EvaluatedFormatConstraint, right: EvaluatedFormatConstraint
    ) -> EvaluatedFormatConstraint:
        """Evaluates logical (inclusive) or_composition"""

        resulting_format_constraint_fulfilled = left.format_constraint_fulfilled or right.format_constraint_fulfilled
        error_message = FormatErrorMessageExpressionBuilder(left).lor(right).get_expression()

        return EvaluatedFormatConstraint(
            format_constraint_fulfilled=resulting_format_constraint_fulfilled, error_message=error_message
        )

    def xor_composition(
        self, left: EvaluatedFormatConstraint, right: EvaluatedFormatConstraint
    ) -> EvaluatedFormatConstraint:
        """Evaluates exclusive xor_composition"""

        resulting_format_constraint_fulfilled = left.format_constraint_fulfilled ^ right.format_constraint_fulfilled
        error_message = FormatErrorMessageExpressionBuilder(left).xor(right).get_expression()

        return EvaluatedFormatConstraint(
            format_constraint_fulfilled=resulting_format_constraint_fulfilled, error_message=error_message
        )


def evaluate_format_constraint_tree(
    parsed_tree: Tree, input_values: Mapping[str, EvaluatedFormatConstraint]
) -> EvaluatedFormatConstraint:
    """
    Evaluates the tree built from the format constraint expressions with the help of the FormatConstraintTransformer.

    :param parsed_tree: Tree
    :param input_values: dict(condition_key, EvaluatedFormatConstraint)
        :return: EvaluatedFormatConstraint
    """
    if not all(isinstance(input_value, EvaluatedFormatConstraint) for input_value in input_values.values()):
        raise ValueError(f"Please make sure that the passed values are {EvaluatedFormatConstraint.__name__}s.")
    try:
        result = FormatConstraintTransformer(input_values).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc

    return result


async def format_constraint_evaluation(
    format_constraints_expression: Optional[str], entered_input: str
) -> FormatConstraintEvaluationResult:
    """
    Evaluation of the format constraint expression.
    """
    error_message: Optional[str] = None
    format_constraints_fulfilled: bool
    if not format_constraints_expression:
        format_constraints_fulfilled = True
    else:
        parsed_tree_fc: Tree = parse_condition_expression_to_tree(format_constraints_expression)
        all_evaluatable_format_constraint_keys: List[str] = [
            t.value for t in parsed_tree_fc.scan_values(lambda v: isinstance(v, Token))  # type:ignore[attr-defined]
        ]
        input_values: Dict[str, EvaluatedFormatConstraint] = await _build_evaluated_format_constraint_nodes(
            all_evaluatable_format_constraint_keys, entered_input
        )
        resulting_evaluated_format_constraint_node: EvaluatedFormatConstraint = evaluate_format_constraint_tree(
            parsed_tree_fc, input_values
        )
        format_constraints_fulfilled = resulting_evaluated_format_constraint_node.format_constraint_fulfilled
        error_message = resulting_evaluated_format_constraint_node.error_message  # pylint:disable=no-member

    return FormatConstraintEvaluationResult(
        format_constraints_fulfilled=format_constraints_fulfilled, error_message=error_message
    )


async def _build_evaluated_format_constraint_nodes(
    evaluatable_format_constraint_keys: List[str], entered_input: str
) -> Dict[str, EvaluatedFormatConstraint]:
    """Build evaluated format constraint nodes."""

    evaluator: FcEvaluator = inject.instance(FcEvaluator)
    evaluated_format_constraints = await evaluator.evaluate_format_constraints(
        evaluatable_format_constraint_keys, entered_input
    )
    return evaluated_format_constraints



""" Tests for the parsing of the conditions tests (Mussfeldprüfung) """

import json
from typing import List

import pytest  # type:ignore[import]

from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtract, CategorizedKeyExtractSchema
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.expressions.condition_expression_parser import extract_categorized_keys
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint

pytestmark = pytest.mark.asyncio


class TestCategorizedKeyExtraction:
    @pytest.mark.parametrize(
        "expression, expected_key_extract",
        [
            pytest.param(
                "[1]",
                CategorizedKeyExtract(
                    hint_keys=[], requirement_constraint_keys=["1"], format_constraint_keys=[], package_keys=[]
                ),
            ),
            pytest.param(
                "([1] U [2]) O ([1] X [3])",
                CategorizedKeyExtract(
                    hint_keys=[],
                    requirement_constraint_keys=["1", "2", "3"],
                    format_constraint_keys=[],
                    package_keys=[],
                ),
            ),
            pytest.param(
                "[100]U([2]U([53]O[4]))[999][502]",
                CategorizedKeyExtract(
                    hint_keys=["502"],
                    requirement_constraint_keys=["2", "4", "53", "100"],
                    format_constraint_keys=["999"],
                    package_keys=[],
                ),
            ),
            pytest.param(
                "[100]U([2]U([53]O[4]))[999][502]U[123P]",
                CategorizedKeyExtract(
                    hint_keys=["502"],
                    requirement_constraint_keys=["2", "4", "53", "100"],
                    format_constraint_keys=["999"],
                    package_keys=["123P"],
                ),
            ),
        ],
    )
    async def test_extraction_of_categorized_keys_from_condition_expression(
        self, expression: str, expected_key_extract: CategorizedKeyExtract
    ):
        """
        Tests that the CategorizedKeyExtract is generated correctly.
        """
        actual = await extract_categorized_keys(expression)
        assert actual == expected_key_extract

    @pytest.mark.parametrize(
        "key_extract, expected_cers",
        [
            pytest.param(
                CategorizedKeyExtract(
                    hint_keys=[], requirement_constraint_keys=[], format_constraint_keys=[], package_keys=[]
                ),
                [],
                id="0 FC, 0 RC",
            ),
            pytest.param(
                CategorizedKeyExtract(
                    hint_keys=[], requirement_constraint_keys=["1"], format_constraint_keys=[], package_keys=[]
                ),
                [
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={},
                        requirement_constraints={"1": ConditionFulfilledValue.FULFILLED},
                        packages={},
                    ),
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={},
                        requirement_constraints={"1": ConditionFulfilledValue.UNFULFILLED},
                        packages={},
                    ),
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={},
                        requirement_constraints={"1": ConditionFulfilledValue.UNKNOWN},
                        packages={},
                    ),
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={},
                        requirement_constraints={"1": ConditionFulfilledValue.NEUTRAL},
                        packages={},
                    ),
                ],
                id="0 FC, 1 RC",
            ),
            pytest.param(
                CategorizedKeyExtract(
                    hint_keys=[], requirement_constraint_keys=[], format_constraint_keys=["901"], package_keys=[]
                ),
                [
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={"901": EvaluatedFormatConstraint(format_constraint_fulfilled=True)},
                        requirement_constraints={},
                        packages={},
                    ),
                    ContentEvaluationResult(
                        hints={},
                        format_constraints={"901": EvaluatedFormatConstraint(format_constraint_fulfilled=False)},
                        requirement_constraints={},
                        packages={},
                    ),
                ],
                id="1 FC, 0 RC",
            ),
        ],
    )
    def test_possible_cer_generation_small_results(
        self, key_extract: CategorizedKeyExtract, expected_cers: List[ContentEvaluationResult]
    ):
        actual = key_extract.generate_possible_content_evaluation_results()
        # We only test the small edge cases as real code.
        # This quickly gets super large. 2 FCs * 2 RCs is already 64 results
        assert actual == expected_cers

    # pylint:disable=fixme
    # todo: register as custom mark https://docs.pytest.org/en/stable/mark.html
    ALL_LARGE_TEST_CASES = pytest.mark.datafiles(
        "./unittests/content_evaluation_result_generation/example0.json",
        "./unittests/content_evaluation_result_generation/example1.json",
    )

    @pytest.mark.parametrize(
        "test_file_path",
        [
            pytest.param("example0.json", id="0 FC, 3 RC"),
            pytest.param("example1.json", id="2 FC, 3 RC"),
        ],
    )
    @ALL_LARGE_TEST_CASES
    def test_possible_cer_generation_large_results(self, test_file_path, datafiles):
        file_content = json.load(datafiles / test_file_path)
        categoried_keys = CategorizedKeyExtractSchema().load(file_content["categorizedKeyExtract"])
        expected_result = ContentEvaluationResultSchema(many=True).load(file_content["expected_result"])
        actual = categoried_keys.generate_possible_content_evaluation_results()
        # json_string = ContentEvaluationResultSchema(many=True).dumps(actual)
        assert actual == expected_result
